// C Header File
// Created 19/09/02; 17:37:26

//--------------------------------------------------------------------------
// <tsr.h>
// The resident routines that will be installed. Some routines are used
// before being installed.

// Ok, this is quite dirty, because GCC could reorganize the variables and
// the functions, but it seems to work.
// And -mpcrel can't be used, because we would  use it for the whole program,
// but some 'bsr.l' are generated that GNU AS do not like with this switch
// activated. And we can't EX_patch the routine, the offsets table is for the
// whole program, not only for the TSR.
// The problem of relocation may come out during the access of global
// variables. Always read the GNU assembly source file generated to see if
// there are not accesses to these variables with absolute addressing mode.
// Note that since tsr.h is included before any instruction in install.c, we
// can use the same strings in tsr.h and in install.c and install_sub.h :
// the string will be created between asm ("NewEvHook:"); and 
// asm ("EndNewEvHook:"); .
// And the global variables used by the TSR must be defined with asm() and
// 'extern', to be sure that they are defined between these 2 marks.
// 
// Any ideas how to share properly functions with the installation routine
// and the TSR ?
//--------------------------------------------------------------------------


#include "macros.h"
#include "functions.h"
#include "structures.h"
#include "install_sub.h"


asm ("NewEvHook:");     // to mark the limits of the event hook
// The header for Kevin Kofler's convention :
asm (".ascii \""EV_HOOK_NAME"\"");
asm ("OldEvHook:\n .long 0");
asm ("jmp AppExeHandler(%pc)\n");    // run the event handler and quit to the AMS

extern EVENT_HANDLER OldEvHook;
extern char *DTfolders_Ptr;
// A pointer to the first folder of the Desktop Folders Structure
asm ("DTfolders_Ptr:\n .long 0");
// The array for the numerical shortcuts : index+2 is the corresponding key
// to press to run the App. Each internal App name is 8 + sizeof('\0') bytes
// long, and is a null-terminated string.
// Shortcut '1' is for 'Refresh Desktop', and is not in the array.
extern char Shortcuts[8][9];
asm (
"Shortcuts:\n"
".ascii \"TIEQUED\"\n"
".space 2\n"
".ascii \"TIWINDED\"\n"
".space 1\n"
".ascii \"TIGRAPH\"\n"
".space 2\n"
".ascii \"TITABLED\"\n"
".space 1\n"
".ascii \"TIDMED\"\n"
".space 3\n"
".ascii \"TIPRGMED\"\n"
".space 1\n"
".ascii \"TITEXTED\"\n"
".space 1\n"
".ascii \"TIINSLVR\"\n"
".space 1\n"
);
// TRUE if AppExeHandler() is being executed, to avoid handling the  events
// if the handler is called recursively.
// Initialized to FALSE.
extern BOOL HandlingEvent;
asm ("HandlingEvent:\n .word 0");

														
//==========================================================================														
// The entry point of the event hook.
// It must be defined with attribute (stkparm) beacause we have activated
// the GCC switch -mregparm, and the AMS is giving the event pointer on
// the stack.
//==========================================================================
void __attribute ((__stkparm__)) AppExeHandler (EVENT *event) {
	// We need to do this to avoid that GCC reads and writes to HandlingEvent
	// with an absolute address.
	BOOL *volatile HandlingEventPtr = &HandlingEvent;
	
	// AppExeHandler() can be called recursively because we are using
	// EV_startApp() and we are running TI-BASIC programs.
	// This may pose problems while handling the shortcuts.
	// So do nothing if AppExeHandler() is already running.
	if (*HandlingEventPtr == TRUE) {
		if (*OldEvHook)
			(*OldEvHook)(event);     // run the old event hook handler
		return;
	}
	*HandlingEventPtr = TRUE;

	// To break, and execute the last line of the function (the execution of
	// the old event handler) -> do while (0)
	do {
	if (event->Type == CM_KEYPRESS) {
		unsigned short keycode = event->extra.Key.Code;
		unsigned short keymod = event->extra.Key.Mod;
		unsigned short kdiamond = KEY_DIAMOND;
		AppID DesktopID;
		
		// remove the diamond modifier, to test even the shortcuts for the kbdprgms
		keycode &= ~kdiamond;
		if (   ((keycode >= '1' && keycode <= '9') || keycode == 'a')
		    // if we are not in a window ('Edit Categories' or the clock setup window)
		    && (!event->RunningApp)
		    // and if the App holding the focus is the Desktop
		    && ((DesktopID = custom_EV_getAppID (DESKTOP_INTERNAL_NAME)) == (AppID)EV_currentApp)) {
	    //=====================================================================
			// The user is running a built-in App, Refresh Desktop with a numerical
			// shortcut, or a kdbprgm.
	    //=====================================================================
			AppID AppToRun;
			custom_StartTypes StartType;
			
			if (keycode == '1' && keymod == 0) {
				RefreshDesktop ();
				// We cannot use 'pushkey(KEY_APPS)' to run the Desktop again, to refresh
				// its display, because it is already running !
				// The parameter for EV_startApp() is c_AP_START_NEW instead of AP_START_CURRENT
				// to force the Desktop to run again.
				// Note that the App selected in the Desktop after this will change to the
				// last App (or fake App) run, but it is not really a problem.
				EV_startApp (DesktopID, c_AP_START_NEW);
				break;     // quit the hook
			}
			
			if (keymod == kdiamond) {
				// We are running a kbdprgm.
				RunApplication (keycode - '0');
				break;
			}
			
			// Else it's a numerical shortcut for a built-in App.
			StartType = c_AP_START_CURRENT;
			if (keycode == 'a' && keymod == 0)
				AppToRun = custom_EV_getAppID ("TIHOME");
			
			else {
				AppToRun = custom_EV_getAppID (Shortcuts[keycode - '2']);
				if ((APP_Flags)OO_GetAppAttr (AppToRun, OO_APP_FLAGS) & APP_CON) {
					HANDLE PopupHandle = H_NULL;
				
					// The App has a Current/Open/New sub-menu, we must draw it to
					// get start type from the user.
					if (   !(PopupHandle = PopupNew (OO_GetAppAttr (AppToRun, OO_APP_NAME), 0))
					    || !PopupAddText (PopupHandle, -1, XR_stringPtr (XR_Current), 0)
					    || !PopupAddText (PopupHandle, -1, XR_stringPtr (XR_Open), 0)
					    || !PopupAddText (PopupHandle, -1, XR_stringPtr (XR_New), 0)) {
						ST_helpMsg ("Not enough memory.");
						if (PopupHandle)
							HeapFree (PopupHandle);
						pushkey (KEY_APPS);    // it will run the Desktop app again
						break;
					}
					PrepareScreen ();
					StartType = PopupDo (PopupHandle, CENTER, CENTER, 0);
					
					if (!StartType) {
						// Escape was pressed.
						// We cannot use 'pushkey(KEY_APPS)' to run the Desktop again,
						// because it is already running !
						// The parameter for EV_startApp() is c_AP_START_NEW instead of
						// AP_START_CURRENT to force the Desktop to run again.
						// Note that the App selected in the Desktop after this will change
						// to the last App (or fake App) run, but it is not really a problem.
						EV_startApp (DesktopID, c_AP_START_NEW);
						break;     // quit the hook
					}
					StartType = (StartType == 1 ? c_AP_START_CURRENT : StartType == 2 ? c_AP_START_OPEN : c_AP_START_NEW );
					HeapFree (PopupHandle);		
				}     // if ((APP_Flags)OO_GetAppAttr ...
			}     // else (<not TIHOME>)
				
			EV_startApp (AppToRun, StartType);
			break;     // do not run the Desktop app again
		}     // if (((keycode >= '1' ...
	}     // if (event->Type == CM_KEYPRESS)
	
	//========
	// Else...
	//========
	else if (event->Type == CM_STARTTASK)    // Is the AMS launching an app ?
		RunApplication (0);
		
	} while (0);

	*HandlingEventPtr = FALSE;
	if (*OldEvHook)
		(*OldEvHook)(event);     // run the old event hook handler
}     // AppExeHandler()



//==========================================================================
// Run a real Flash App, a fake one or a kbdprgm from the Desktop.
// If KbdPrgm is not 0, we are running the kbdprgm number KbdPrgm.
// The App Desktop will not be run at the end of the function if a kbdprgm
// is run.
// The hook must be quitted immediatly after this function is executed.
//==========================================================================
void RunApplication (unsigned short KbdPrgm) {
	ACB *AppACBPtr;
	char *FileName, *PathPtr;
	SYM_ENTRY *FileSymPtr;
	// where the TI-Basic command to execute the program will be stored
	// -> folder\name()\0
	char SimpleCommand[8+1+8+2+1];
	// where the TI-Basic command to execute the program with ttstart or
	// sstart will be stored (if it is a PPG file) 
	// -> xstart_folder\xstart("folder\name")\0
	char XStartCommand[8+1+7+2+8+1+8+2+1];
	HANDLE CommandHandle;
	ESI *TopEstackPtr;
	ESI TopEstackBackup;
	ESQ PrgmType;
	char KbdPrgmBuffer[8];     // for it's name, without the '\0' at the end

	TopEstackPtr = &top_estack;
	TopEstackBackup = *TopEstackPtr;

	if (KbdPrgm) {
		// Create the name of the kbdprgm file to run.
		// FileName will point to the buffer instead of the database.
		strcpy (KbdPrgmBuffer, "kbdprgm");
		KbdPrgmBuffer[7] = (char)KbdPrgm + '0';
		FileName = KbdPrgmBuffer;
	}
	
	else {
		AppACBPtr = (ACB*)HeapDeref (EV_currentApp);
		// Is this app 'Refresh Desktop' ?
		if (AppACBPtr->publicstorage == ACB_REFRESH_DT_MAGIC) {
			// prepare the screen, to avoid the display of a destroyed Desktop
			PrepareScreen ();
			// refresh the Desktop (with a message)
			RefreshDesktop ();
			pushkey (KEY_APPS);    // it will run the Desktop app again 
			return;
		}
		// Is this app a fake app ?
		if ((AppACBPtr -> publicstorage != ACB_PRGM_MAGIC)  \
		 && (AppACBPtr -> publicstorage != ACB_TXT_MAGIC))
			return;     // no : do nothing
	
		PrepareScreen();
		// point to the description, after the object frame and the icon 
		FileName = (char*)(AppACBPtr->appData) + sizeof(OO_Hdr)
								+ (sizeof(OO_Attr) * 3) + ICON_SIZE;
		while (*FileName++)     // search the end of the description
			continue;
		// in the data file, there was perhaps a second null character ending the
		// description, to avoid address errors. This null character has been copied
		// here. So if it exists, jump it. 
		if (!(*FileName))
			FileName++;
		// FileName points now to the name of the program to run or the text file to view

	}     // else (<!KbdPrgm>)		
	
	//But FileName is not a string !
	// We use XStartCommand[] as a buffer to convert it to a string
	// (XStartCommand[] is not used for the moment)
	memcpy (XStartCommand, FileName, 8);
	XStartCommand[8] = 0;
		
	// Now we can use XStartCommand as the 'filename' parameter
	// Search for PPG/TI-BASIC/ASM/TEXT
	if (!(FileSymPtr = FindFile (XStartCommand, NULL, TRUE, TRUE, SimpleCommand, NULL))) {
		// the file was not found
		if (!KbdPrgm) {
			// Display a message and run the Desktop only if we are not running a
			// kbdprgm.
			ST_helpMsg ("File not found. Run 'Refresh Desktop'.");
			pushkey (KEY_APPS);    // it will run the Desktop app again
		}
			return;
	}

	PathPtr = SimpleCommand;
	while (*PathPtr)     // search the end of the folder name
		PathPtr++;
	// now it points to the \0
	*PathPtr++ = '\\';     // the '\' between the folder and the name
	memcpy (PathPtr, FileName, 8);     // append the name of the file
	*(PathPtr + 8) = 0;

	PrgmType = *HToESI (FileSymPtr->handle);
	if (PrgmType != ASM_TAG && PrgmType != FUNC_TAG) {
		// then it's a PPG file or a TEXT file
		if (PrgmType == TEXT_TAG) {
			// Search if the text viewer is still on the calc.
			// Search for ASM and TI-BASIC, and get its folder name in XStartCommand.
			if (!FindFile (TXT_VIEWER_NAME, NULL, FALSE, FALSE, XStartCommand, NULL)) {
				ST_helpMsg ("Error: '"TXT_VIEWER_NAME"' not found...");
				if (!KbdPrgm)
					pushkey (KEY_APPS);    // it will run the Desktop app again
				return;
			}
			// append the name of the viewer after the folder name
			strcat (XStartCommand, "\\"TXT_VIEWER_NAME"(\"");
		}
			
		else {    // if (PrgmType == <PPG>)
			// First check if the Flash App Super Start is installed.
			if (custom_EV_getAppID ("sstart")) {
				strcpy (XStartCommand, "sstart(\"");				
			}
			
			else {
				// Else try to find find the path of ttstart().
				// Search for ASM (and TI-BASIC, but it's not really a problem)
				if (!FindFile ("ttstart", NULL, FALSE, FALSE, XStartCommand, NULL)) {
					ST_helpMsg ("Error: 'ttstart' not found...");
					if (!KbdPrgm)
						pushkey (KEY_APPS);    // it will run the Desktop app again 
					return;
				}
				// Append the name of the launcher for PPG files after the folder name.
				strcat (XStartCommand, "\\ttstart(\"");
			}
		}   // else -> if (PrgmType == <PPG>)

		// Append the path of the file to run with ttstart/sstart or to view
		// with the text viewer.
		strcat (XStartCommand, SimpleCommand);
		strcat (XStartCommand, "\")");     // -> append ")	
		push_parse_text (XStartCommand);    // we will run this command
	}     // if (PrgmType != ASM_TAG ...
	
	else {
		// it's an ASM file : we just need to had '()' at the end of the command
		strcat (SimpleCommand, "()");
		push_parse_text (SimpleCommand);    // we will run this command
	}
		
	// We cannot use HS_popEStack() in a TSR, because it would pop ALL the estack !
	// (even what the AMS has pushed before).(Thank you Kevin Kofler for having
	// fixed this)
	// So we are using Samuel Stearley's trick : we copy just what has been added
	// to the stack in a block of memory.
	{
		unsigned short ParsedDataSize = *TopEstackPtr - TopEstackBackup;
		unsigned short *NewBlockPtr;
		
		// + 2 for the size at the beginning of the block
		if (!(CommandHandle = HeapAllocHigh (ParsedDataSize + 2))) {
			ST_helpMsg ("Not enough memory.");
			if (!KbdPrgm)
				pushkey (KEY_APPS);    // it will run the Desktop app again
			return;
		}
		NewBlockPtr = HeapDeref (CommandHandle);
		// NG_execute needs this
		*NewBlockPtr++ = ParsedDataSize;
		// TopEstackBackup + 1 points to the parsed data
		memcpy (NewBlockPtr, TopEstackBackup + 1, ParsedDataSize);
		*TopEstackPtr = TopEstackBackup;     // restore the top_estack pointer
	}     // GetParseData block

	// We must catch the errors to avoid getting in a loop of EV_defaultHandler()
	// when an error is thrown during the parse of the tokenized string or during
	// the execution of the program
	TRY
		// NG_execute() can be used because HW220xTSR is installed
		NG_execute (CommandHandle, 0);
	ONERR
		// if an error is thrown, just display the error message
		ERD_dialog (errCode, FALSE);
	ENDTRY	
	HeapFree (CommandHandle);
	
	if (!KbdPrgm)
		pushkey (KEY_APPS);    // it will run the Desktop app again
}     // RunApplication ()



//==========================================================================
// Prepares the screen when running a program from the Desktop, to avoid the
// display of a destroyed Desktop App.
//==========================================================================
void PrepareScreen (void) {
	unsigned short i;
	char *ScrPtr;
	
	ClearScreen();		
	// redraw the horizontal line over the status line, erased by ClearScreen()
	// point to the left of the line
	ScrPtr = LCD_MEM + 30 * (TI89 ? (100-7) : (128-7));
	for (i = 30; i; i--)     // 30 bytes width
		*ScrPtr++ = 0xFF;     // draw 8 pixels
	// redraw the status line erased by ClearScreen()
  // (ST_eraseHelp() works only if a message is being displayed)
	ST_helpMsg("");
	ST_eraseHelp();
}     // PrepareScreen()



//==========================================================================
// The fake app used to refresh the Desktop.
// It will be called like a function by AppExeHandler.
// It appears at the first place in the folder 'All' of the Desktop.
//==========================================================================

void RefreshDesktop (void) {
	HANDLE FilesList_handle;
	unsigned short FilesNumber, DeletedFilesNumber;
	char *FilesListPtr_top;
	HANDLE DeletedFilesList_handle;

	//=====================================================================
	// We scan the linked list of the ACBs and search the fake apps.
	// If a fake app does not exist any more, we add its AppID to the
	// DeletedFilesList, and we delete its AppID and its object frame.
	// If it still exists, we check the null word just after its name in the
	// FilesList, by writing TRUE.
	//=====================================================================
	
	{
		unsigned short i;
		short found = 0;     // (initialized to avoid a warning)
		char *FilesListPtr, *FileName;	
		AppID CurrentACB;
		ACB *CurrentACB_Ptr;
		BOOL TextViewerFound;
		FileOfFilesListStruct OneFileSt; // for sizeof()
		AppID *DeletedFilesListPtr; // always points to the last element of the list,
																// where we can store the AppID of the next deleted
																// file found.
		
		TextViewerFound = FindTxtViewer();
		if (!(FilesList_handle = GetSupportedFilesList (TextViewerFound, &FilesNumber))) {
			ST_helpMsg ("Not enough memory.");
			return;
		}
		FilesListPtr_top = HeapDeref (FilesList_handle);
		
		// create an empty list that will contain the AppID of the files that
		// have been deleted
		// (there is always enough memory to create a block of sizeof(AppID) bytes)
		DeletedFilesListPtr = HeapDeref (DeletedFilesList_handle = HeapAlloc (sizeof(AppID)));
		DeletedFilesNumber = 0;     // there are 0 files in the list
		
		CurrentACB = OO_firstACB;	
		do {
			__label__ AddToDeletedFilesList;
			
			CurrentACB_Ptr = HeapDeref (CurrentACB);
			if (CurrentACB_Ptr->publicstorage != ACB_PRGM_MAGIC \
			 && CurrentACB_Ptr->publicstorage != ACB_TXT_MAGIC)
				// Then it's not a fake app. Deal with the next ACB.
				continue;
				
			if (CurrentACB_Ptr->publicstorage == ACB_TXT_MAGIC && !TextViewerFound)
				// it's a text file and there is not viewer : add it to DeletedPrgmsList
				goto AddToDeletedFilesList;
			// else point to the description, after the object frame and the icon 
			FileName = (char*)(CurrentACB_Ptr->appData) + sizeof(OO_Hdr)
								+ (sizeof(OO_Attr) * 3) + ICON_SIZE;
			while (*FileName++)     // search the end of the description
				continue;
			// in the data file, there was perhaps a second null character ending the
			// description, to avoid address errors. This null character has been copied
			// here. So if it exists, jump it. 
			if (!(*FileName))
				FileName++;
			// FileName points now to the name of the file

			// search if the file is still on the calc
			for (FilesListPtr = FilesListPtr_top, i = FilesNumber; i; i--) {
				FileOfFilesListStruct OneFile;
				if (!(found = memcmp (FilesListPtr, FileName, sizeof (OneFile.name))))
					break;
				FilesListPtr += sizeof (OneFile);     // next file in the list
			}		
			
			if (found) {     // -> not found
			 AddToDeletedFilesList:
				// The file is not on the calc : it has been deleted or renamed
				// (or it's a text file and the text viewer was not found).
				// Add it's AppID to DeletedFilesList.
				*DeletedFilesListPtr = CurrentACB;
				DeletedFilesNumber++;     // one more file in the list
	
				// resize the list to create space for the next AppID that may be added
				if(!(HeapRealloc (DeletedFilesList_handle, (DeletedFilesNumber<<1) + sizeof (AppID)))) {
					HeapFree (DeletedFilesList_handle);
					HeapFree (FilesList_handle);
					ST_helpMsg ("Not enough memory.");
					return;
				}

				// make the pointer point to the last AppID (that does not exist actually)
				DeletedFilesListPtr  = HeapDeref (DeletedFilesList_handle) + (DeletedFilesNumber<<1);
				// Be careful : CurrentACB_Ptr may not be valid any more since the HeapRealloc()
				// refresh it
				CurrentACB_Ptr = HeapDeref (CurrentACB);
				// delete the object frame, and remove the ACB from the linked list
				RemoveACBAndOFrame (CurrentACB_Ptr);				
				// We have to free CurrentACB.
				// But be careful : the OO_NextACB (CurrentACB) in the while loop condition
				// will not work.
				// So we set CurrentACB to the previous ACB, whose 'next' member now
				// equals OO_NextACB(CurrentACB)
					AppID ACB_to_delete = CurrentACB;
					CurrentACB = CurrentACB_Ptr->prev;
					HeapFree (ACB_to_delete);	

			}     // if (found)
	
			else
				// The file still exists.
				// Check it in the files list, just after its name (TwoBytes.action).
				*(FLIST_ACTION*)(FilesListPtr + sizeof (OneFileSt.name)) = FLIST_EXISTS;
			
		} while ((CurrentACB = OO_NextACB (CurrentACB))); // do

	}     // check_and_delete block
	
	
	//=====================================================================
	// Now we remove from the Desktop folders all the files that have
	// been deleted
	//=====================================================================

	{
		AppID *DeletedFilesPtr;
		unsigned short i;

	  // for all the files deleted
	  DeletedFilesPtr = HeapDeref (DeletedFilesList_handle);
		for (i = DeletedFilesNumber; i; i--, DeletedFilesPtr++)
			// remove *DeletedFilesPtr from all the folders
			RemoveFromFolders (TRUE, *DeletedFilesPtr);
		
	}     // clean_up_DesktopFolders block
	HeapFree (DeletedFilesList_handle);     // we do not need it any more
	
	
	//=====================================================================
	// Create now ACBs and object frames for the files with FLIST_CREATE
	// or FLIST_CREATE_TXT after their name in the list (TwoBytes.action) :
	// they are the new files on the calc since the last time
	// RefreshDesktop() was run (they have not been checked with
	// FLIST_EXISTS).
	//=====================================================================
	
	{
		HANDLE ICO_handle;
		char *FirstICO;
		
		{
		// search the file containing the icons & descriptions
			SYM_ENTRY *SymPtr;
			if(!(SymPtr = FindFile (ICO_FILENAME, ICO_TYPE, FALSE, FALSE, NULL, NULL))) {
				HeapFree (FilesList_handle);
				ST_helpMsg ("Error : '"ICO_FILENAME"' not found.");
				return;
			}	
			ICO_handle = SymPtr->handle;
		}
			
		// to keep the pointers on the icon file valid even if we create blocks
		// of memory, causing a Heap Compression
		// don't forget to unlock it !
		if ((FirstICO = HLock (ICO_handle)) == NULL) {
			HeapFree (FilesList_handle);
			ST_helpMsg ("Not enough memory.");
			return;
		}
		// jump the size of the file
		FirstICO += 2;
		
		if (*(((unsigned long*)FirstICO)++) != ICO_MAGIC) {   // not the right file
			HeapUnlock (ICO_handle);
			HeapFree (FilesList_handle);
			ST_helpMsg ("Error : '"ICO_FILENAME"' not found.");
			return;
		}
		
		if (!CreateACBs (FirstICO, FilesListPtr_top, FilesNumber)) {
			// not enough memory
			HeapUnlock (ICO_handle);
			HeapFree (FilesList_handle);
			ST_helpMsg ("Not enough memory.");	
			return;	
		}
		
	}     // icons_for_new_files block
	HeapFree (FilesList_handle);     // we do not need it any more


	//=====================================================================
	// Create the backup of the folders of the Desktop
	//=====================================================================
	
	{
		SYM_ENTRY *SymPtr;
		HANDLE folder_handle;
		char bck_file_folder_name[9];
		BOOL custom_folder = FALSE; // TRUE if the backup is saved in a special folder
		HANDLE Backup_handle;     // handle of the new backup

		// First we delete the previous backup file if it exists
		if ((SymPtr = FindFile (BCK_FILENAME, BACKUP_TYPE, FALSE, FALSE, bck_file_folder_name,
				 &folder_handle))) {
			__label__ CannotDelete;
			char *Bck_Ptr;
			HSym bck_hsym = MakeHSym (folder_handle, SymPtr);
		
			Bck_Ptr = HeapDeref (SymPtr->handle);
			Bck_Ptr += 2;     // jump the size of the file
			if (*((long*)Bck_Ptr)++ == BACKUP_MAGIC)
				// It is a backup. The new backup will be saved in the folder
				// given by bck_file_folder_name[]
				custom_folder = TRUE;
			// Even if it is not a backup, delete the old file (the user shouldn't have
			// a file named BCK_FILENAME with an extension BACKUP_TYPE that is not a
			// backup)
			// be careful : HSymDel() does not work on archived files
			if (SymPtr->flags.bits.archived == 1) {
				if (!EM_moveSymFromExtMem (NULL, bck_hsym))
					goto CannotDelete;
			}
			if (!HSymDel (bck_hsym)) {
			 CannotDelete:
				ST_helpMsg ("Cannot delete the previous backup file.");
				return;
			}
		}     // if ((SymPtr = FindFile ...

		
		// Create the new backup file
		{
			unsigned short i;
			char *BackupPtr, *FoldersPtr;
			AppID *AppsPtr;
			unsigned short FileSize, AppsNumber;
			HANDLE AppsList_handle;
			BackupFileSt BckFile;     // just for sizeof()
			BckFolderStruct BckFolder;
			DT_FolderStruct DT_FolderSt;
	
			// place for the size of the file and the magic number
			if (!(Backup_handle = HeapAlloc (sizeof (unsigned short) + sizeof (BckFile.magic)))) {
				ST_helpMsg ("No memory to create the backup file.");
				return;
			}
			FileSize = sizeof (unsigned short) + sizeof (BckFile.magic);
			// the size will be written later
			BackupPtr = HeapDeref (Backup_handle) + sizeof (unsigned short);
			*(unsigned long*)BackupPtr = BACKUP_MAGIC;
			
			// for all the folders of the Desktop
			// (DTfolders_Ptr += sizeof (DT_FolderSt) to jump to the next folder)
			for (i = DESKTOP_FOLDERS_NUM, FoldersPtr = DTfolders_Ptr; i; i--,
					 FoldersPtr += sizeof (DT_FolderSt)) {
				
				// place for the name of the folder and the number of apps
				if (!HeapRealloc (Backup_handle, FileSize + sizeof (BckFolder.folder_name)
				  + sizeof (BckFolder.programs_number))) {
					HeapFree (Backup_handle);
					ST_helpMsg ("No memory to create the backup file.");
					return;
				}
				BackupPtr = HeapDeref (Backup_handle) + FileSize;
				FileSize += sizeof (BckFolder.folder_name) + sizeof (BckFolder.programs_number);
				// copy the name of the folder
				memcpy (BackupPtr, FoldersPtr, sizeof (BckFolder.folder_name));
				BackupPtr += sizeof (BckFolder.folder_name);
				// get the handle of the list of the apps in the folder
				AppsList_handle = *(HANDLE*)(FoldersPtr + sizeof (DT_FolderSt.folder_name) \
					+ sizeof (DT_FolderSt.must_be_null));
				
				if (!AppsList_handle)
					// there are no files in this folder : write AppsNumber = 0
					AppsNumber = *(unsigned short*)BackupPtr = 0;
				else {
					AppsPtr = HeapDeref (AppsList_handle);
					*(unsigned short*)BackupPtr = AppsNumber = GetAppsNumberInFolder (AppsPtr);
				
					// place to write the BCK_TAG and the name for all the (fake?) apps
					HeapRealloc (Backup_handle, FileSize + \
						(sizeof ( BckFolder.files[0])) * AppsNumber);
					BackupPtr = HeapDeref (Backup_handle) + FileSize;
					FileSize += (sizeof ( BckFolder.files[0])) * AppsNumber;
					AppsPtr = HeapDeref (AppsList_handle);
				
					for (; AppsNumber; AppsNumber--, AppsPtr++) {
						ACB *ACB_Ptr;
						ACB_Ptr = HeapDeref (*AppsPtr);
						unsigned long magic = ACB_Ptr -> publicstorage;
						
						if (magic == ACB_REFRESH_DT_MAGIC)     // the fake app 'Refresh Desktop'
							//write the BCK_TAG. No name is needed.
							*((BCK_TAG*)BackupPtr)++ = BCK_REFRESH;
						
						else if (magic == ACB_PRGM_MAGIC || magic == ACB_TXT_MAGIC) {
							// it's a fake app (but not Refresh Desktop)
							char *FramePtr;
							
							// write the BCK_TAG					
							*((BCK_TAG*)BackupPtr)++ = ACB_PRGM_MAGIC ? BCK_PRGM : BCK_TEXT;
							// for a fake app, the name is after the icon and the description
							// in the object frame block
							// point to the description, after the object frame and the icon 
							FramePtr = (char*)ACB_Ptr->appData + sizeof(OO_Hdr)
								+ (sizeof(OO_Attr) * 3) + ICON_SIZE;
							while (*FramePtr++)     // search the end of the description
								continue;
							// in the data file, there was perhaps a second null character ending
							// the description, to avoid address errors. This null chatacter has
							// been copied here. So if it exists, jump it. 
							if (!(*FramePtr))
								FramePtr++;
							// FramePtr points now to the name of the fake app
							memcpy (BackupPtr, FramePtr, sizeof (BckFolder.files[0].file_name));
						}     // else <if (fake_app)>
						
						else {     // it's a real flash app
							*((BCK_TAG*)BackupPtr)++ = BCK_FLASH_APP;    // write the BCK_TAG
							// search the name of the flash app and write it
							// (+4 to jump the magic number, see '7.1.3. Application Header'
							// in TIFS documentation)
							memcpy (BackupPtr, (char*)(ACB_Ptr -> appHeader) + 4, 8);
						}     // else if (flash_app)
						
						// jump the name we have just written	
						BackupPtr += sizeof (BckFolder.files[0].file_name);					
					}     // for (; AppsNumber; ...	
			
				}     // else <if (!AppsList_handle)>
				
			}     // for (i = DESKTOP_FOLDERS_NUM; ...
			
			// write the file extension
			if (!HeapRealloc (Backup_handle, FileSize + \
			 sizeof (BACKUP_TYPE_HEXA) + sizeof ((unsigned short)OTH_TAG))) {
				HeapFree (Backup_handle);
				ST_helpMsg ("No memory to create the backup file.");
				return;
			}
			BackupPtr = HeapDeref (Backup_handle) + FileSize;
			*((long*)BackupPtr)++ = BACKUP_TYPE_HEXA;
			*(unsigned short*)BackupPtr = (unsigned short)OTH_TAG;
			// write the size of the file
			// be careful : the 2 first bytes must not be taken into account
			FileSize += sizeof (BACKUP_TYPE_HEXA) + sizeof ((unsigned short)OTH_TAG) \
				- sizeof (unsigned short);
			*(unsigned short*)HeapDeref (Backup_handle) = FileSize;
			
		}     // CreateNewBackup block
		

		// Create the entry in the VAT
		{
			// enough space to contain a SymName
			char new_bck_path[1+8+1+sizeof(BCK_FILENAME)+1];
			HSym NewHSym;
			char *SymNamePtr;

			// the first byte must be null in SymName strings
			new_bck_path[0] = 0;
			if (custom_folder)
				// the custom folder is the folder where the previous backup was
				strcpy (new_bck_path + 1, bck_file_folder_name);
			else
				// create it in the folder 'main'
				strcpy (new_bck_path + 1, "main");
			strcat (new_bck_path + 1, "\\"BCK_FILENAME);
			SymNamePtr = new_bck_path + 1 + strlen (new_bck_path + 1);
			NewHSym = SymAdd (SymNamePtr);
			if (!NewHSym.folder) {
				ST_helpMsg ("Cannot create the backup file...");
				HeapFree (Backup_handle);
				return;
			}
			
			DerefSym (NewHSym) -> handle = Backup_handle;		
			// archive the file
			if (!EM_moveSymToExtMem (NULL, NewHSym)) {
				ST_helpMsg ("Cannot archive the backup file...");
				return;
			}
					
		}     // CreateVATEntry block
		
	}     // CreateFoldersBackup block
	
		ST_helpMsg ("Desktop refreshed.");
}     // RefreshDesktop()



//==========================================================================
// Scans all the folders of the Desktop.
// If action is TRUE, searches AppIDToRemove in the folders and remove it
// from all the folders (except the folder 'All')
// If action is FALSE, Searches in all the folders (including the folder 
// 'All' !) all the fake apps, and removes them from the folders. Parameter
// AppIDToRemove is then ignored. 
//==========================================================================
void RemoveFromFolders (BOOL action, AppID AppIDToRemove) {
	unsigned short i, j;
	char *FoldersPtr;     // points to the Desktop folders structure
	AppID *AppsPtr;     // points to the AppID of the apps in the folders
	unsigned short FolderSize;     // size of the list of the Apps in the folder
	HANDLE AppsListHandle;     // handle of the currently scanned Apps list
	DT_FolderStruct FolderSt;     // just for sizeof()
	unsigned short AppsIndex; // index of the current AppID in a folder
														 // starting with sizeof(AppID), incremented by sizeof(AppID)

	// from the first folder after "All", pointing on the apps list handle
	FoldersPtr = DTfolders_Ptr + sizeof (FolderSt.folder_name) + sizeof (FolderSt.must_be_null);
	// for all the folders of the Desktop
 	i = DESKTOP_FOLDERS_NUM;
 	// if action is FALSE, we must also scan the folder 'All'
	if (!action) {
		FoldersPtr -= sizeof (FolderSt);
		i++;     // scan one more folder
	}

	// (FoldersPtr += sizeof (FolderSt) to jump to the handle of the next files list)
	for (; i; i--, FoldersPtr += sizeof (FolderSt)) {

		AppsListHandle = *(HANDLE*)FoldersPtr;
		if (AppsListHandle == H_NULL)
			continue;     // there is no files in this folder, deal with the next one
		AppsPtr = HeapDeref (AppsListHandle);
		FolderSize = GetAppsListSize (AppsPtr);
		
		// scan the list of the apps (the list is null terminated)
		for (AppsIndex = sizeof(AppID), j = 0; *AppsPtr;) {

			// if we are searching for AppIDToRemove and it has been found,
			// or if we are searching all the fake apps and we have found one (or
			// if we have found 'Refresh Desktop')
			if ((action && *AppsPtr == AppIDToRemove) ||
					((!action && ((ACB*)HeapDeref (*AppsPtr))->publicstorage == ACB_PRGM_MAGIC)
					|| (((ACB*)HeapDeref (*AppsPtr))->publicstorage == ACB_REFRESH_DT_MAGIC))) {
				// To remove the AppID from the list, we copy the end of the list
				// 2 bytes to the left (overwritting this AppID)
				// be careful : there is overlap, we use memmove.
				memmove (AppsPtr, AppsPtr+1, FolderSize - AppsIndex);
				// if there is still Apps in the folder
				if (FolderSize != (sizeof(AppID)<<1)) {
					// there is one AppID less in the folder
					HeapRealloc (AppsListHandle, FolderSize - sizeof(AppID));
					// correct the size of the folder
					FolderSize -= sizeof(AppID);
					// we must get the new AppsPtr, which is now invalid
					AppsPtr = (AppID*)HeapDeref (AppsListHandle) + j;
				}     // if (FolderSize != ...
				else {
					// there are no more files in the folder :
					// (there were only one app and the terminating null word)
					// erase the list and write that it does not exist any more
					HeapFree (AppsListHandle);
					*(AppID*)FoldersPtr = H_NULL;
					break;    // deal with the next folder
				}     // else <if (FolderSize != ... >
			}     // if ((action && *AppsPtr == AppIDToRemove ...
			else {
				AppsIndex += sizeof(AppID);
				AppsPtr++;
				j++;
			}     // else <// if ((action && ...>
			
		}     // 	for (AppsIndex = 0; *AppsPtr; ...			
	}     // for (i = DESKTOP_FOLDERS_NUM; ...		

}     // RemoveFromFolders()



//==========================================================================
// Remove the object frame associated with the ACB poitned to by ACB_Ptr,
// and remove the ACB from the linked list of the ACBs.
// Do NOT free the ACB.
//==========================================================================
void RemoveACBAndOFrame (ACB *ACB_Ptr) {
	// delete the object frame of the app
	HeapFreePtr ((void*)(ACB_Ptr->appData));
	// Delete the ACB and remove it from the linked list only if the object
	// frame exists (FreeCreatedACB_and_Frames() could have been run if there
	// wasn't enough memory to create the frame)
	// CurrentACB_Ptr->prev cannot be H_NULL (the fake apps are always
	// installed after the built-in apps)
	((ACB*)HeapDeref (ACB_Ptr->prev))->next = ACB_Ptr->next;
	// but CurrentACB_Ptr->next can be null
	if (ACB_Ptr->next)
		((ACB*)HeapDeref (ACB_Ptr->next))->prev = ACB_Ptr->prev;
}     // RemoveACBAndOFrame()



//==========================================================================
// Creates a list of the name of all the ASM, PPG, TI-BASIC programs of the
// calc.
// If SearchTXT is TRUE, TEXT files are also listed.
// Each name is 8 bytes long (unused bytes are initialized to 0), ans each
// name is followed by 2 bytes initialized to FLIST_CREATE for ASM/PPG/TI-BASIC,
// and to FLIST_CREATE_TXT fot TEXT files).
// These two bytes will be used in the TSR to refresh the Desktop, and by
// CreateACBs().
// If there are 2 files with the same name in different folders, the
// name is listed only once.
// GetSupportedFilesList() returns the handle of the block of memory containing
// the list, and writes the number of files in the list in the variable
// given in argument by refererence.
// If no files were found, FilesNumber = 0, and the returned value is
// the handle of a 4 bytes block containing unknown data.
// If there wasn't enough memory to create the list, the returned handle is
// H_NULL, and FilesNumber = 0.
// The created block of memory is locked to avoid problems with Heap Compression
// when the lists is used.
// Don't forget to free the created block after use.
//==========================================================================
HANDLE GetSupportedFilesList (BOOL SearchTXT, unsigned short *FilesNumber) {
	HANDLE list_handle;
	SYM_ENTRY *SymPtr;
	char *FilePtr;
	unsigned short j, current_size;
	unsigned short FilesNum = 0;     // number of files added so far in the list
	ESQ FileType;
	FileOfFilesListStruct FileStruct;     // used only for sizeof()
	
	SymPtr = SymFindFirst (NULL, 2);     	// Scan all the folders and all the files
	current_size = 0;
	list_handle = HeapAlloc (2);     // just to create a block of memory...
	
	if (!list_handle)
		return list_handle;     // return H_NULL

	for (; SymPtr != NULL; SymPtr = SymFindNext()) {
		
		if (SymPtr->flags.bits.folder == 1)     // it's a folder
			continue;
		FilePtr = HToESI (SymPtr->handle);
		FileType = *(ESQ*)FilePtr;
		if (FileType != ASM_TAG && FileType != FUNC_TAG
		    && !(SearchTXT && FileType == TEXT_TAG)) {
			// if it is not an ASM or a TI-BASIC prgm, and not a TEXT file if we are
			// also searching for TEXT files, is it a PPG file ?
			if (*(FilePtr)   != (char)OTH_TAG ||
  				*(FilePtr-4) != 'p'  ||
  				*(FilePtr-3) != 'p'  ||
  				*(FilePtr-2) != 'g')
  					continue;
  	}     // if (*FilePtr != ...
  	
  	// check if there is not already a program with the same name in the list
		FilePtr = HeapDeref (list_handle);
  	for (j = FilesNum; j; j--) {
  		if (!memcmp (SymPtr->name, FilePtr, sizeof (FileStruct.name)))
  		  // the file is in the list
  			break;
  		FilePtr += sizeof (FileStruct);
  	}
  	if (j) // the previous for-loop has been broken : the same name was found
  		continue;
  	
  	// Now we can had the name of the file to the list :
  	// enlarges the block
  	HeapRealloc (list_handle, current_size += sizeof (FileStruct));
  	// copies the name at the right place in the block
  	memcpy (FilePtr = HeapDeref(list_handle) + FilesNum * sizeof (FileStruct), \
  		SymPtr->name, sizeof (FileStruct.name));
		// the 2 bytes
			*(FLIST_ACTION*)(FilePtr + sizeof (FileStruct.name)) \
				 = (FileType == TEXT_TAG) ? FLIST_CREATE_TXT : FLIST_CREATE;
		(FilesNum)++;     // one more file in the list
		
	}     // for (; SymPtr != NULL; ...

	*FilesNumber = FilesNum;     // return the number of files in the list	
	return HeapLock (list_handle);
}     // GetSupportedFilesList()


//==========================================================================
// Searches the first file called <filename> in all the folders of the calc.
// If type is NULL, we are searching a file of custom type PPG or of type ASM
// or TI-BASIC.
// If type is NULL and SearchPPG is FALSE, only ASM and TI-BASIC programs will
// be searched.
// If type is NULL and SearchTXT is TRUE, TEXT files will be also searched.
// If type is not NULL, we are searching a file of custom type <type>
// (SearchPPG be read).
// If folder_name is not NULL, the name of the folder where the file is (if a
// file is found) is copied to 'folder_name' as a null terminated string.
// If folder_handle is not NULL, the handle of the VAT of the folder is
// stored in the variable pointed to by folder_handle (if a file is found).
// Returns NULL if no matching file was found.
//==========================================================================
SYM_ENTRY *FindFile (const char *filename, const char *type, BOOL SearchPPG,
										 BOOL SearchTXT, char *folder_name, HANDLE *folder_handle) {
	SYM_ENTRY *SymPtr;
	char *FilePtr;
	BOOL custom_type = type ? TRUE : FALSE;     // TRUE if we are searching a custom type

	// Scan all the folders and all the files
	SymPtr = SymFindFirst (NULL, 2);
	for ( ;SymPtr != NULL; SymPtr = SymFindNext()) {
		
		if (SymPtr->flags.bits.folder == 1) {     // it's a folder
			if (folder_name)
				strcpy (folder_name, SymPtr->name);     // store its name
				// -> if a matching file is found, folder_name will point to the name
				// where the file is.
			if (folder_handle)
				*folder_handle = SymPtr->handle;     // the handle of the VAT of this folder
			continue;     // deal with the files of this folder	
		}     // if (SymPtr->flags.bits.folder == 1)
		
		if (strcmp (SymPtr->name, filename))     // not the right name
			continue;
		FilePtr = (char*)HToESI (SymPtr->handle);
		if ((!custom_type)
				// is it an ASM program, a USER DEF program, or a text file (if SearchTXT is true) ?
				&& ((*FilePtr == (char)ASM_TAG) || (*FilePtr == (char)FUNC_TAG)
				 || (SearchTXT && (*FilePtr == (char)TEXT_TAG))))  
					break;     // yes, we have found a matching file

		// if custom_type==FALSE, the following code will be used to check if it
		// a ppg file
		if (*FilePtr != (char)OTH_TAG)     // not a custom type
			continue;
		if (*(--FilePtr))    // there must be a char(0) before the tag
			continue;
		while (*(--FilePtr))     // search the previous (char)0
			continue;
		FilePtr++;     // point to the custom type string
		if (custom_type) {
			if (!strcmp (FilePtr, type))     // we have found it !
				break;
		}
		else     // we are not searching a custom type : we are searching a ppg file
			if (SearchPPG && !strcmp (FilePtr, "ppg"))
				break;
	
	}     // for ( ;SymPtr != NULL;  ...
	return SymPtr;
} // FindFile()


//==========================================================================
// Each fake flash app will have this object frame.
// The two NULL pointers need to be replaced by pointers to the description
// and to the icon of the program or text file.
// A pointer to this object frame pattern is &FramePattern.
//==========================================================================

FRAME(FramePattern, OO_SYSTEM_FRAME, 0, OO_APP_FLAGS, 3)
   ATTR(OO_APP_FLAGS, APP_INTERACTIVE)
   ATTR(OO_APP_NAME, NULL)
   ATTR(OO_APP_ICON, NULL)
ENDFRAME


//==========================================================================
// The icon for text files (a BITMAP of 22x22)
//==========================================================================

static BITMAP TxtIcon = {
	22, 22,
	{	0x36,0xda,0x00,0x5b,0x6d,0x00,0xa4,0x93,0x88,0xa4,0x93,0x40,
		0x80,0x01,0x5c,0x80,0x01,0x48,0xbc,0x5d,0x48,0x80,0x01,0x48,
		0xbc,0x01,0x48,0x80,0x01,0x40,0x80,0x01,0x54,0x80,0x01,0x54,
		0xbf,0xfd,0x48,0x80,0x01,0x54,0xbf,0xfd,0x54,0x80,0x01,0x40,
		0xbf,0xfd,0x5c,0x80,0x01,0x48,0xbe,0x01,0x48,0x80,0x01,0x48,
		0x7f,0xfe,0x88,0x00,0x00,0x00
	}
};


//==========================================================================
// Create an ACB and on object frame for one program or text file.
// If CreateTXT is TRUE, an ACB for a text file is created : TxtIcon is used
// and the description of the file is "<filename>.txt"
// If CreateTXT is TRUE, ICO_Ptr and IcoDataSize are not used.
// Else, ICO_Ptr points to the icon+description, IcoDataSize is the size of
// the icon+description, Name points to the name of the file (8 bytes),
// LastACB_ID is the handle of the last ACB in the linked list, NewACB_ID_Ptr
// is a pointer to a variable where the handle of new ACB will be stored
// (if NewACB_ID_Ptr is NULL, the handle will not be stored in the variable),
// IsRDT must be TRUE if the ACB to create is for the fake app Refresh
// Desktop (it is used to select the right magic number).
// Returns false in case of a memory error.
//==========================================================================
BOOL CreateOneACB (BOOL CreateTXT, const void *ICO_Ptr,
									 unsigned short IcoDataSize, const char *Name,
									 AppID LastACB_ID, AppID *NewACB_ID_Ptr, BOOL IsRDT) {

  ACB *NewACB_Ptr;
  char *NewFrame_Ptr, *FramePtr;
  AppID NewACB_ID;
	char NameStr[9];     // if CreateTXT is TRUE
  unsigned long TxtDataSize = 0;     // initialization to avoid a warning
  
  if (!(NewACB_ID = HeapAlloc (sizeof(ACB))))
		return FALSE;
  
  NewACB_Ptr = HeapDeref (NewACB_ID);
  // installed, and built-in to avoid being displayed in the VAR-LINK screen
  NewACB_Ptr->flags = ACB_INSTALLED | ACB_BUILTIN;
  NewACB_Ptr->publicstorage = IsRDT ? ACB_REFRESH_DT_MAGIC : ACB_PRGM_MAGIC;
  // we set AppHeader+0x24 to the low word of the auto-int1 vector which 
  // contains necessarily an even word (-> no language localization)
  NewACB_Ptr->appHeader = (void*)(0x64+2-0x24);
  NewACB_Ptr->myID = NewACB_ID;
  NewACB_Ptr->prev = LastACB_ID;     // the previous last ACB of the linked list
	((ACB*)HeapDeref (LastACB_ID))->next = NewACB_ID; //link our ACB to the list
	// in case of a memory error, FreeCreatedACB_and_Frames() must know
	// where to stop its scanning :
  NewACB_Ptr->next = H_NULL;
  
  // We have to create the frame block containing :
  // the frame header (sizeof(OO_Hdr))
  // 3 object attributes (3*sizeof(OO_Attr))
  // the icon and the description (ICODataSize or
  //   ICON_SIZE + strlen (Name) + TXT_EXTENSION_SIZE)
  // the name of the file (8 bytes)
  if (CreateTXT) {
  	// We need to get the length of Name, but it's not a string !
		memcpy (NameStr, Name, 8);
		NameStr[8] = '\0';
  	TxtDataSize = ICON_SIZE + strlen (NameStr) + TXT_EXTENSION_SIZE;	
	}
  if (!(NewFrame_Ptr = HeapAllocPtr (sizeof(OO_Hdr) + (sizeof(OO_Attr) * 3) \
   + (CreateTXT ? TxtDataSize : IcoDataSize) + 8)))
		return FALSE;
	// Be careful : NewACB_Ptr may become invalid after HeapAllocPtr(),
  // we refind it with HeapDeref().
  ((ACB*)HeapDeref (NewACB_ID)) -> appData = (pFrame)NewFrame_Ptr;
  
  	// copy the frame pattern and the attributes that we will
  	// next fill up
  	memcpy (NewFrame_Ptr, &FramePattern, sizeof(OO_Hdr) + (sizeof(OO_Attr) * 3));
		FramePtr = NewFrame_Ptr + sizeof(OO_Hdr) + (sizeof(OO_Attr) * 3);
		
		// copy just after the frame pattern the icon and the description
		if (!CreateTXT) {
			// ICO_Ptr is pointing the data we have to copy
			memcpy (FramePtr, ICO_Ptr, IcoDataSize);
			FramePtr += IcoDataSize;
		}
		else {     // if (CreateTXT)
			// we use TxtIcon
			memcpy (FramePtr, &TxtIcon, ICON_SIZE);
			FramePtr += ICON_SIZE;
			strcat (strcpy (FramePtr, NameStr), TXT_EXTENSION);
			FramePtr += TxtDataSize - ICON_SIZE;     // point after all this stuff
		}
		
		// copy the name of the file after all this stuff
		// this name is in the list of ASM/PPG/BASIC/TEXT files
		memcpy (FramePtr, Name, 8);

	// The pointer to the icon (the second NULL in FramePattern) is filled
	// up with the adress of the icon (which is just after FramePattern)
  *(void**)(NewFrame_Ptr + sizeof(OO_Hdr) + (sizeof(OO_Attr)<<1) + 4) = \
  		NewFrame_Ptr + sizeof(OO_Hdr) + (sizeof(OO_Attr) * 3);
   // The pointer to the description (the first NULL) is also filled up
	// (the description is just after the icon) :
	 *(void**)(NewFrame_Ptr + sizeof(OO_Hdr) + sizeof(OO_Attr) + 4) = \
  		NewFrame_Ptr + sizeof(OO_Hdr) + (sizeof(OO_Attr) * 3) + ICON_SIZE;
	
	if (NewACB_ID_Ptr)     // store it only if it is not NULL
		*NewACB_ID_Ptr = NewACB_ID;
	return TRUE;
}     // CreateOneACB()


//==========================================================================
// Returns the handle of the last ACB of the linked list
//==========================================================================
AppID GetLastACB (void) {
	AppID NextACB_ID, LastACB_ID;
	LastACB_ID = OO_firstACB;
	while ((NextACB_ID = OO_NextACB (LastACB_ID)))
		LastACB_ID = NextACB_ID;
	return LastACB_ID;
}     // GetLastACB()


//==========================================================================
// Create an ACB and an object frame for all the files in the list starting
// at FilesListPtr (created with GetSupportedFilesList()).
// FilesNumber is the number of files in this list.
// FirstICO points to the file containing the icons (just after the magic
// number).
// Only ACBs for the files with FLIST_CREATE or FLIST_CREATE_TXT after
// their names in the list (the element TwoBytes.action of FileOfFilesListStruct)
// are created.
// If action is FLIST_CREATE_TXT, the icon TxtIcon will be used, else it will
// be the icon in the data base (if it is found).
// If the ACB for a file is created, the AppID of its ACB is written after
// its name in the list (in the element TwoBytes.CreatedACB of
// FileOfFilesListStruct).
// Else, TwoBytes.Created is set to FALSE.
// Returns FALSE if there was not enough memory.
//==========================================================================
BOOL CreateACBs (void *FirstICO, void *FilesListPtr, unsigned short FilesNumber) {
	char *ICO_Ptr;
	unsigned short i, j, IconsNumber, IcoDataSize ;
	AppID LastACB_ID, NewACB_ID;
	FileOfFilesListStruct OneFileSt;     // just for sizeof()
	DataBaseStruct DataBaseEntry;
	short found = 1;     // = 0 if the file was found in the data base

	IconsNumber = *(((unsigned short*)FirstICO)++);
	// and FirstICO points now to the data of the first program in the 'icon' file
	LastACB_ID = GetLastACB();
	
	// we search for each files in the list if we can find its icon+description
	// (FilesListPtr += sizeof (OneFileSt) to jump to the next file)
	for (i = FilesNumber; i; i--, FilesListPtr += sizeof (OneFileSt)) {
		// create ACBs only if TwoBytes.action is not FLIST_EXISTS
		if (*(FLIST_ACTION*)(FilesListPtr + sizeof (OneFileSt.name)) == FLIST_EXISTS)
			continue;
		ICO_Ptr = FirstICO;

		// We are going to scan the data base only if the file is not a text file.
		// If it is a text file we can create the ACB immediately.
		if (*(FLIST_ACTION*)(FilesListPtr + sizeof (OneFileSt.name)) == FLIST_CREATE_TXT) {
			if (!CreateOneACB (TRUE, NULL, 0, FilesListPtr, LastACB_ID, \
			 &NewACB_ID, FALSE))
				return FALSE;
			// write the handle of the new ACB where the FLIST was (TwoBytes.CreatedACB)
 			*(AppID*)(FilesListPtr + sizeof (OneFileSt.name)) = NewACB_ID;
 			// now it's the last ACB of the linked list
 			LastACB_ID = NewACB_ID;
			continue;     // deal with the next file in the list
		}

		// Else it's not a TEXT file. Scan all the data base.
		for (j = IconsNumber; j; j--) {
		
			if ((found = memcmp (ICO_Ptr, FilesListPtr, sizeof (OneFileSt.name)))) {
 				// not the right name
				ICO_Ptr += sizeof (DataBaseEntry.name);     // jump the name
				// point to the next data block
				unsigned short DataSize = *((unsigned short*)ICO_Ptr)++;
				ICO_Ptr += DataSize;  
				continue;
			}

			// else :
			ICO_Ptr += sizeof (DataBaseEntry.name);   // point to the size of the data
			// we have found the icon+description of the program :
  		// create the ACB for this program
			IcoDataSize = *((unsigned short*)ICO_Ptr)++;
			if (!CreateOneACB (FALSE, ICO_Ptr, IcoDataSize, FilesListPtr, LastACB_ID,
					&NewACB_ID, FALSE))
				return FALSE;
  		// We must fill the 2 bytes after the file name in the files list
  		// with the handle of the corresponding ACB we have created
  		// (the list with the ACB handles will be used later)
  		*(AppID*)(FilesListPtr + sizeof (OneFileSt.name)) = NewACB_ID;
  		LastACB_ID = NewACB_ID;     // now it's the last ACB of the linked list
			// We have found the icon and the description. Now deal with the next
			// file in the list
			break;
		
		}     // for (j = IconsNumber; ...
			
		if (found)
		  // -> not found
			// Then all the data base has been scanned, and the program was not found.
			// (we will enter this 'if' only if the above 'break' was NOT executed)
			// Write FALSE to TwoBytes.Created.
			*(BOOL*)(FilesListPtr + sizeof (OneFileSt.name)) = FALSE;
		
	}     // for(i = FilesNumber; ...
	
	return (TRUE);     // creation successful
}     // CreateACBs


//--------------------------------------------------------------------------

asm ("EndNewEvHook:");     // to mark the limits of the event hook
